package mvd.jester.tests;

import java.math.RoundingMode;
import java.util.Map.Entry;
import com.google.common.math.LongMath;
import mvd.jester.model.Segment;
import mvd.jester.model.Task;
import mvd.jester.model.SystemSetup;

/**
 * MaiaBertogna
 */
public class MaiaBertogna extends AbstractTest {

    private final SystemSetup systemSetup;

    public MaiaBertogna(SystemSetup systemSetup) {
        this.systemSetup = systemSetup;
    }

    @Override
    public boolean runSchedulabilityCheck() {
        responseTimes.clear();
        for (Task t : systemSetup.getTasks()) {
            responseTimes.put(t, calculateResponseTime(t));
        }

        for (Entry<Task, Long> keyValuePair : responseTimes.entrySet()) {
            if (keyValuePair.getKey().getDeadline() < keyValuePair.getValue().longValue()) {
                return false;
            }
        }

        return true;
    }

    @Override
    public String getName() {
        return "MaiaBertogna";
    }

    private long calculateResponseTime(Task task) {
        long minimumWcet = getMinimumWcet(task);
        long responseTime = minimumWcet;
        long previousResponseTime = 0;

        do {
            previousResponseTime = responseTime;
            long interference = 0;

            for (Task t : systemSetup.getTasks()) {
                if (t.getPeriod() < task.getPeriod()) {
                    long maxNumberOfJobsOfT = t.getMaximumParallelism();
                    for (int p = 0; p < maxNumberOfJobsOfT; ++p) {
                        interference += Math.min(getTaskInterference(t, responseTime, p + 1),
                                responseTime - minimumWcet + 1);
                    }
                }
            }

            double taskInterference = (double) interference / systemSetup.getNumberOfProcessors();
            long selfInterference = 0;

            long maxNumberOfJobs = task.getMaximumParallelism();
            for (int p = 0; p < maxNumberOfJobs; ++p) {
                selfInterference +=
                        Math.min(getSelfInterference(task, p + 1), responseTime - minimumWcet + 1);
            }

            long totalInterference = (long) Math.floor(taskInterference + selfInterference);

            responseTime = minimumWcet + totalInterference;
        } while (previousResponseTime != responseTime);


        return responseTime;
    }

    private long getSelfInterference(Task task, long parallelism) {
        long interference = 0;

        for (Segment s : task.getSegments()) {
            if (s.getNumberOfJobs() >= parallelism + 1) {
                s.getJobWcet();
            }
        }

        return interference;
    }

    private long getTaskInterference(Task task, long interval, long parallelism) {
        long responseTime = responseTimes.get(task);
        long minWcet = getMinimumWcet(task);
        long period = task.getPeriod();
        long numberOfJobs =
                (LongMath.divide(interval + responseTime - minWcet, period, RoundingMode.FLOOR)
                        + 1);

        long workload = 0;

        for (Segment s : task.getSegments()) {
            if (s.getNumberOfJobs() >= parallelism) {
                workload += s.getJobWcet();
            }
        }

        long interference = numberOfJobs * workload;

        return interference;
    }

    private long getMinimumWcet(Task task) {
        long minWcet = 0;
        for (Segment s : task.getSegments()) {
            minWcet += s.getJobWcet();
        }

        return minWcet;
    }

}
